import string
import numpy as np

from ase.geometry import cell_to_cellpar

from protosearch import build_bulk

path = build_bulk.__path__[0]
wyckoff_data = path + '/Wyckoff.dat'

crystal_class_coordinates = {'P': [[0, 0, 0]],
                             'A': [[0, 0, 0],
                                   [0, 0.5, 0.5]],
                             'C': [[0, 0, 0],
                                   [0.5, 0.5, 0]],
                             'I': [[0, 0, 0],
                                   [0.5, 0.5, 0.5]],
                             'R': [[0, 0, 0]],
                             'R-3': [[0, 0, 0],
                                     [2/3, 1/3, 1/3],
                                     [1/3, 2/3, 2/3]],
                             'F': [[0, 0, 0],
                                   [0, 0.5, 0.5],
                                   [0.5, 0, 0.5],
                                   [0.5, 0.5, 0]],
                             'H': [[0, 0, 0],
                                   [2/3, 1/3, 1/3],
                                   [1/3, 2/3, 2/3]]}

alphabet = list(string.ascii_lowercase) + ['A', 'B', 'C', 'D']


class WyckoffSymmetries:
    """Class to handle Wyckoff positions and symmetries"""

    def __init__(self,
                 spacegroup,
                 wyckoffs=None,
                 ):
        self.spacegroup = spacegroup

        self.set_wyckoff_symmetries()
        if wyckoffs is not None:
            self.wyckoffs = wyckoffs
            self.set_wyckoff_mapping()

    def set_wyckoff_symmetries(self):
        """Construct dictionary with  """

        self.wyckoff_symmetries = {}
        self.wyckoff_site_symmetries = {}
        self.wyckoff_multiplicities = {}
        position = None
        with open(wyckoff_data, 'r') as f:
            i_sg = np.inf
            i_w = np.inf
            for i, line in enumerate(f):
                if '1 {} '.format(self.spacegroup) in line \
                   and not i_sg < np.inf:
                    i_sg = i
                    SG_letter = line.split(' ')[2]
                    if 'R-3' in SG_letter or 'R3' in SG_letter:
                        SG_letter = 'R-3'
                    else:
                        SG_letter = SG_letter[0]
                    self.class_coordinates = crystal_class_coordinates[SG_letter]

                if i > i_sg:
                    if len(line) == 1:
                        break
                    if line.split(' ')[1] in alphabet:
                        i_w = i
                        position, site_symmetry = line.split(' ')[1:3]
                        self.wyckoff_multiplicities.update(
                            {position: int(line.split(' ')[0])})
                        self.wyckoff_site_symmetries.update(
                            {position: site_symmetry})
                        wyckoff_array = np.zeros([0, 3, 4])
                    if i > i_w:
                        if len(line) < 40:
                            break
                        arr = np.array(line.split(' ')[:-1], dtype=float)
                        for cc in self.class_coordinates:
                            x = arr[:4].copy()
                            x[3] += cc[0]
                            y = arr[4:8].copy()
                            y[3] += cc[1]
                            z = arr[8:].copy()
                            z[3] += cc[2]
                            new_arr = np.expand_dims(
                                np.array([x, y, z]), axis=0)

                            wyckoff_array = np.append(
                                wyckoff_array, new_arr, axis=0)
                if position:
                    self.wyckoff_symmetries.update({position: wyckoff_array})

    def get_free_wyckoffs(self):
        free_wyckoffs = []
        for w_letter in sorted(self.wyckoff_symmetries.keys()):
            w_array = self.wyckoff_symmetries[w_letter][0]
            M = w_array[:, :3]
            if not np.all(M == np.zeros([3, 3])):
                free_wyckoffs += [w_letter]

        return free_wyckoffs

    def set_wyckoff_mapping(self):
        # Name of coordinate variables (Enumerator consistent)
        self.coor_variables = []
        # Map atomic indices to those of same wyckoff position
        self.symmetry_map = []
        # Symmetry matrix and offset for each atom
        self.wyckoff_map = []
        # Atom index free to move or now
        self.free_atoms = []
        # Wyckoff position for atom indices
        self.atoms_wyckoffs = []
        # free Wyckoff directions
        self.w_free_dir = {}

        xyz = list('xyz')
        relative_positions = []

        symbols = []
        atoms_index = 0

        for i_w, w in enumerate(self.wyckoffs):
            w_id = w + str(i_w)
            m = self.wyckoff_multiplicities[w]
            w_arrays = self.wyckoff_symmetries[w]
            self.w_free_dir.update({w_id: []})

            for w_array in w_arrays:
                self.wyckoff_map += [w_array]
                M = w_array[:, :3]

                free_wyckoff = False
                for d in range(3):
                    if not np.all(M[:, d] == 0):
                        v = xyz[d] + w_id
                        if not v in self.coor_variables:
                            self.coor_variables += [v]
                        self.w_free_dir[w_id] += [d]
                        free_wyckoff = True
                self.free_atoms += [free_wyckoff]
                self.atoms_wyckoffs += [w_id]

            sym_indices = [atoms_index + i for i in range(m)]
            for i in sym_indices:
                self.symmetry_map += [sym_indices]
            atoms_index += m

    def get_prototype_name(self, species):
        alphabet = list(string.ascii_uppercase)
        unique_symbols = []
        symbol_count = []

        for i, s in enumerate(species):
            if s in unique_symbols:
                index = unique_symbols.index(s)
                symbol_count[index] += self.wyckoff_multiplicities[self.wyckoffs[i]]
            else:
                symbol_count += [self.wyckoff_multiplicities[self.wyckoffs[i]]]
                unique_symbols += [s]

        min_rep = min(symbol_count)

        for n in list(range(1, min_rep + 1))[::-1]:
            if np.all(np.array(symbol_count) % n == 0):
                repetition = n
                break

        p_name = ''
        for ii, i in enumerate(np.argsort(symbol_count)):
            p_name += alphabet[ii]
            factor = symbol_count[i] // n
            if factor > 1:
                p_name += str(factor)

        p_name += '_' + str(repetition)  # // len(self.class_coordinates))

        added_species = ['']
        for i, w in enumerate(self.wyckoffs):
            s = species[i]

            if not s == added_species[-1]:
                p_name += '_'
                p_name += w
            else:
                if w == p_name[-1]:
                    p_name += '2'
                elif p_name[-1].isdigit() and w == p_name[-2]:
                    p_name = p_name[:-1] + str(int(p_name[-1]) + 1)
                else:
                    p_name += w

            added_species += [s]
        p_name += '_' + str(self.spacegroup)

        return p_name

    def get_wyckoff_transform_vector(self, vector, index_a, index_b, cell):
        vector = np.dot(np.linalg.inv(cell.T),
                        vector)

        vector0 = vector.copy()
        normv0 = np.linalg.norm(vector0)
        M_a = self.wyckoff_map[index_a][:3, :][:, :3]
        M_b = self.wyckoff_map[index_b][:3, :][:, :3]
        red_dim_x = 0
        red_dim_y = 0
        for i in range(3):
            i_x = i - red_dim_x
            i_y = i - red_dim_y
            if np.all(M_a[i_x, :] == 0):
                M_a = np.delete(M_a, i_x, 0)
                vector = np.delete(vector, i_x, 0)
                red_dim_x += 1
            if np.all(M_a[:, i_y] == 0):
                M_a = np.delete(M_a, i_y, 1)
                M_b = np.delete(M_b, i_y, 1)
                red_dim_y += 1

        M_a_inv = np.dot(np.linalg.inv(np.dot(M_a.T, M_a)), M_a.T)
        transform_vector = np.dot(M_b, np.dot(M_a_inv, vector))

        if not list(transform_vector):
            transform_vector = np.zeros(3)

        transform_vector = np.dot(cell.T, transform_vector)

        return transform_vector

    def get_inverse_wyckoff_matrix(self, M):
        x = []
        y = []
        for i in range(3):
            if not np.all(M[i, :] == 0):
                x += [i]
            if not np.all(M[:, i] == 0):
                y += [i]
        M = M[x, :][:, y]

        M_inv = np.dot(np.linalg.inv(np.dot(M.T, M)), M.T)

        return M_inv, x, y

    def get_cell_parameters(self, atoms):
        cell = atoms.cell
        cell_parameters = {}

        parameters = cell_to_cellpar(cell)

        for i, param in enumerate(['a', 'b', 'c',
                                   'alpha', 'beta', 'gamma']):

            cell_parameters.update({param: parameters[i]})

        if np.all(self.free_atoms == 0):
            return cell_parameters

        dir_map = ['x', 'y', 'z']

        relative_positions = np.dot(np.linalg.inv(
            cell.T), atoms.positions.T).T

        unique_atoms_w, indices = np.unique(self.atoms_wyckoffs,
                                            return_index=True)

        for ai in indices:
            if not self.free_atoms[ai]:
                continue
            aw = self.atoms_wyckoffs[ai]
            position = relative_positions[ai]
            w_position = self.get_parameter_from_position(position, aw[0])
            for i, w in enumerate(w_position):
                if np.isclose(w, 0):
                    continue
                w_p_name = dir_map[i] + aw
                cell_parameters.update({w_p_name: w})

        return cell_parameters

    def get_parameter_from_position(self, position, w_position):
        for w_sym in self.wyckoff_symmetries[w_position]:
            match, w_pos = self.is_position_wyckoff(position, w_sym,
                                                    return_coor=True)
            if match:
                return wrap_coordinate(w_pos, 0.5)
        print('position not found: {}:{}'.format(position, w_position))

    def is_position_wyckoff(self, position, w_sym, return_coor=False, tol=1e-3):
        M = w_sym[:, :3]
        c = w_sym[:, 3]
        M_inv, dim_x, dim_y = self.get_inverse_wyckoff_matrix(M)
        r_vec = (position - c)[dim_x]
        for plane in [0, 0.5, -0.5]:
            r_vec_temp = wrap_coordinate(r_vec, plane=plane)

            w_position = np.zeros([3])
            w_position[dim_y] = np.dot(r_vec_temp, M_inv.T)

            test_position = np.dot(w_position, M.T) + c
            test_position = wrap_coordinate(test_position,
                                            plane=plane)

            c_position = wrap_coordinate(position,
                                         plane=plane)
            #print(test_position, c_position)
            if np.all(np.isclose(test_position, c_position, rtol=tol)):
                # print('OK')
                if return_coor:
                    return True, w_position
                else:
                    return True
            # else:
            #    print('No')
        if return_coor:
            return False, None
        else:
            return False


def wrap_coordinate(coor, plane=0.5):
    # Center values in vector arround

    offset = plane + 0.5

    coor = [c - 1 if np.isclose(c, offset) else -c if np.isclose(-c, offset)
            else c - 1 if c > offset else c + 1 if c < offset-1 else c for c in coor]

    coor = [0 if c == -1 else 0 if c == 1 else c for c in coor]

    return coor


def get_wyckoff_letters(spacegroups):

    spacegroup_letters = []
    with open(wyckoff_data, 'r') as f:
        i_sg = np.inf
        sg = 1
        while sg < max(spacegroups):
            for i, line in enumerate(f):
                if '1 {} '.format(sg) in line:
                    i_sg = i
                    spacegroup_letters += [[]]
                if i > i_sg:
                    if i == i_sg + 1:
                        sg += 1
                    if len(line) == 1 or len(line) < 6:
                        continue
                    if line.split(' ')[1] in alphabet:
                        position, site_symmetry = line.split(' ')[1:3]
                        if not position in spacegroup_letters[-1]:
                            spacegroup_letters[-1] += [position]

    return spacegroups, spacegroup_letters



import sys
import shutil
import os
import numpy as np
import tempfile
import unittest
from ase.db import connect

from protosearch.workflow.workflow import Workflow
from protosearch.workflow.standard_states import StandardStates


class WorkflowTest(unittest.TestCase):
    def setUp(self):
        self.pwd = os.getcwd()
        self.tempdir = tempfile.mkdtemp()
        os.chdir(self.tempdir)

    def tearDown(self):
        os.chdir(self.pwd)
        shutil.rmtree(self.tempdir)

    def test_workflow_submit(self):
        WF = Workflow(db_filename='test.db')
        prototype = {'spacegroup': 221,
                     'wyckoffs': ['a', 'd'],
                     'species': ['Ru', 'O']}
        WF.submit(prototype)

    def test_standard_states(self, elements=['K', 'Al']):
        WF = Workflow(db_filename='test.db')
        WF.submit_standard_states(elements=elements)

    def test_job_status(self):
        WF = Workflow(db_filename='test.db')
        WF.check_submissions()


if __name__ == '__main__':
    unittest.main()



from catlearn.regression.gaussian_process import GaussianProcess
ml_methods_dict = {
    "catlearn": "CatLearnGP"}


def get_regression_model(name):
    if name == 'catlearn':
        from protosearch.ml_modelling.regression_model import CatLearnGP as Model
    return Model


class CatLearnGP:

    # CatLearn GP interface

    def __init__(self,
                 features,
                 targets,
                 kernel_type='gaussian',
                 kernel_width=3,
                 kernel_dimension='features',
                 regularization=3e-2,
                 optimize_hyperparameters=True,
                 scale_data=True):

        self.features = features
        self.targets = targets

        kernel_list = [{'type': kernel_type,
                        'width': kernel_width,
                        'dimension': kernel_dimension}]

        self.GP = GaussianProcess(
            train_fp=features,
            train_target=targets,
            kernel_list=kernel_list,
            regularization=regularization,
            optimize_hyperparameters=optimize_hyperparameters,
            scale_data=scale_data
        )
        #print('hyperparam:', self.GP.theta_opt)

    def predict(self, test_features, uncertainty=True):
        prediction = self.GP.predict(test_features,
                                     uncertainty=uncertainty)

        return prediction



import sys
import numpy as np
from spglib import get_symmetry_dataset, standardize_cell
from ase import Atoms
from ase.spacegroup import Spacegroup
from ase.build import cut


from .wyckoff_symmetries import WyckoffSymmetries, wrap_coordinate

from protosearch import build_bulk
path = build_bulk.__path__[0]

wyckoff_data = path + '/Wyckoff.dat'
wyckoff_pairs = path + '/Symmetric_Wyckoff_Pairs.dat'


class PrototypeClassification(WyckoffSymmetries):
    """Prototype classification of atomic structure in ASE Atoms format"""

    def __init__(self, atoms, tolerance=1e-3):

        self.spglibdata = get_symmetry_dataset((atoms.get_cell(),
                                                atoms.get_scaled_positions(),
                                                atoms.get_atomic_numbers()),
                                               symprec=tolerance)

        self.tolerance = tolerance

        self.spacegroup = self.spglibdata['number']
        self.Spacegroup = Spacegroup(self.spacegroup)
        self.atoms = self.get_conventional_atoms(atoms)

        super().__init__(spacegroup=self.spacegroup)

        self.set_wyckoff_species()
        WyckoffSymmetries.wyckoffs = self.wyckoffs
        WyckoffSymmetries.species = self.species


    def get_conventional_atoms(self, atoms):
        """Transform from primitive to conventional cell"""


        std_cell = self.spglibdata['std_lattice']

        positions = self.spglibdata['std_positions']
        numbers = self.spglibdata['std_types']

        atoms = Atoms(numbers=numbers,
                      cell=std_cell,
                      pbc=True)

        atoms.set_scaled_positions(positions)
        atoms.wrap()

        return atoms

    def get_primitive_atoms(self, atoms):
        """Transform from primitive to conventional cell"""

        lattice, scaled_positions, numbers = standardize_cell(atoms,
                                                              to_primitive=True,
                                                              no_idealize=False,
                                                              symprec=1e-5)

        atoms = Atoms(numbers=numbers,
                      cell=lattice,
                      pbc=True)

        atoms.set_scaled_positions(scaled_positions)

        atoms.wrap()

        return atoms

    def set_wyckoff_species(self):
        self.wyckoffs = []
        self.species = []

        relative_positions = np.dot(np.linalg.inv(
            self.atoms.cell.T), self.atoms.positions.T).T

        relative_positions = np.round(relative_positions, 10)

        normalized_sites = \
            self.Spacegroup.symmetry_normalised_sites(
                np.array(relative_positions,
                         ndmin=2))

        equivalent_sites = []
        for i, site in enumerate(normalized_sites):
            indices = np.all(np.isclose(site, normalized_sites[:i+1],
                                        rtol=self.tolerance), axis=1)
            index = np.min(np.where(indices)[0])
            equivalent_sites += [index]

        unique_site_indices = list(set(equivalent_sites))

        unique_sites = normalized_sites[unique_site_indices]
        count_sites = [list(equivalent_sites).count(i)
                       for i in unique_site_indices]

        symbols = self.atoms[unique_site_indices].get_chemical_symbols()

        for i, position in enumerate(unique_sites):
            found = False
            for w in sorted(self.wyckoff_symmetries.keys()):
                m = self.wyckoff_multiplicities[w]
                if not count_sites[i] == m:
                    continue
                for w_sym in self.wyckoff_symmetries[w]:
                    if found:
                        break
                    if self.is_position_wyckoff(position, w_sym):
                        found = True
                        self.wyckoffs += [w]
                        self.species += [symbols[i]]
                        break
            if not found:
                print('Error: position: ', position, ' not identified')

        indices = np.argsort(self.species)

        w_sorted = []
        # for s in set(self.species):
        #    indices = [i for i, s0 in enumerate(self.species) if s0==s]
        #    w_sorted += sorted([self.wyckoffs[i] for i in indices])
        #self.wyckoffs = w_sorted

        free_wyckoffs = self.get_free_wyckoffs()
        self.atoms_wyckoffs = []
        self.free_atoms = []
        for site in equivalent_sites:
            index = unique_site_indices.index(site)
            w_position = self.wyckoffs[index]
            self.atoms_wyckoffs += [w_position + str(index)]
            self.free_atoms += [w_position in free_wyckoffs]

    def get_spacegroup(self):
        return self.spacegroup

    def get_wyckoff_species(self):
        return self.wyckoffs, self.species

    def get_classification(self, include_parameters=True):

        p_name = self.get_prototype_name(self.species)

        structure_name = str(self.spacegroup)
        for spec, wy_spec in zip(self.species, self.wyckoffs):
            structure_name += '_{}_{}'.format(spec, wy_spec)

        prototype = {'p_name': p_name,
                     'structure_name': structure_name,
                     'spacegroup': self.spacegroup,
                     'wyckoffs': self.wyckoffs,
                     'species': self.species}

        if include_parameters:
            cell_parameters = self.get_cell_parameters(self.atoms)

            return prototype, cell_parameters

        else:
            return prototype


def get_wyckoff_pair_symmetry_matrix(spacegroup):
    letters, multiplicity = get_wyckoff_letters_and_multiplicity(spacegroup)

    n_points = len(letters)
    pair_names = []
    for i in range(n_points):
        for j in range(n_points):
            pair_names.append('{}_{}'.format(letters[i], letters[j]))

    M = np.zeros([n_points**2, n_points**2])

    with open(wyckoff_pairs, 'r') as f:
        sg = 1
        for i, line in enumerate(f):
            if len(line) == 1:
                if sg < spacegroup:
                    sg += 1
                    continue
                else:
                    break
            if sg < spacegroup:
                continue
            w_1 = line[:3]
            if not w_1 in pair_names:
                continue
            k = pair_names.index(w_1)
            pairs0 = line.split('\t')[2: -1]
            for w_2 in pairs0:
                j = pair_names.index(w_2)
                M[k, j] = 1

    free_letters = []
    for l in letters:
        i = pair_names.index(l + '_' + l)
        if M[i, i] == 1:
            free_letters += [l]

    np.fill_diagonal(M, 1)

    return pair_names, M, free_letters


def get_wyckoff_letters_and_multiplicity(spacegroup):
    letters = np.array([], dtype=str)
    multiplicity = np.array([])
    with open(wyckoff_data, 'r') as f:
        i_sg = np.inf
        i_w = np.inf
        for i, line in enumerate(f):
            if '1 {} '.format(spacegroup) in line \
               and not i_sg < np.inf:
                i_sg = i
            if i > i_sg:
                if len(line) > 15:
                    continue
                if len(line) == 1:
                    break
                multi, w, sym, sym_multi = line.split(' ')
                letters = np.insert(letters, 0, w)
                multiplicity = np.insert(multiplicity, 0, multi)

    return letters, multiplicity



import io
import numpy as np
from ase.io.vasp import write_vasp

from protosearch.utils.standards import VaspStandards, CommonCalc
from protosearch.utils.valence import VaspValence


class VaspModel:
    def __init__(self,
                 symbols,
                 calc_parameters=None,
                 ncpus=1):

        self.calc_parameters = VaspStandards.calc_parameters
        if calc_parameters is not None:
            self.calc_parameters.update(calc_parameters)
        self.natoms = len(symbols)
        self.all_symbols = symbols
        self.symbols = set(symbols)
        self.ncpus = ncpus
        self.setups = {}
        U_luj = CommonCalc.U_luj
        ldau_luj = {}
        for symbol in symbols:
            if symbol in VaspStandards.paw_potentials:
                self.setups.update({symbol:
                                    VaspStandards.paw_potentials[symbol]})

        if np.any([t in symbols for t in CommonCalc.U_trickers]) and \
           np.any([t in symbols for t in CommonCalc.U_metals]):
            for symbol in symbols:
                ldau_luj.update({symbol:
                                 U_luj.get(symbol,
                                           {'L': 0, 'U': 0, 'J': 0})})

        self.sorted_parameters = VaspStandards.sorted_calc_parameters
        if ldau_luj:
            self.calc_parameters.update({'ldau_luj': ldau_luj})
            self.sorted_parameters = VaspStandards.sorted_calc_parameters + \
                VaspStandards.u_parameters

        self.initial_magmoms = {}
        for symbol in [s for s in symbols if s in CommonCalc.magnetic_trickers]:
            self.initial_magmoms.update(
                {symbol: CommonCalc.initial_magmoms[symbol]})

        nbands = self.calc_parameters['nbands']
        if nbands < 0:
            nbands = self.get_nbands(n_empty=abs(nbands))
        else:
            nbands = self.get_nbands()
        self.calc_parameters.update({'nbands': nbands})

    def get_parameters(self):
        return self.sorted_parameters.copy(), self.sorted_calc_values.copy()

    def get_parameter_dict(self):
        return self.calc_parameters

    def get_model(self):
        """
        Construct model string, which uses the ASE interface
        to Vasp.
        """

        modelstr = get_model_header()
        if self.initial_magmoms:
            modelstr = self.add_initial_magmoms(modelstr)

        modelstr += 'calc = Vasp(\n'

        if self.setups:
            modelstr += '    setups={'
            for symbol, setup in self.setups.items():
                modelstr += "'{}': '{}'".format(symbol, setup)
            modelstr += '},\n'

        for param in self.sorted_parameters:
            value = self.calc_parameters[param]
            if isinstance(value, str):
                modelstr += "    {}='{}',\n".format(param, value)
            elif isinstance(value, dict):
                modelstr += '    {}='.format(param)
                modelstr += '{'
                nkeys = len(value)
                i = 1
                for k, v in value.items():
                    modelstr += "'{}': {}".format(k, v)
                    if i < nkeys:
                        modelstr += ',\n' + ' ' * (len(param) + 6)
                    i += 1
                modelstr += '},\n'
            else:
                modelstr += '    {}={},\n'.format(param, value)

        modelstr += '    )\n\ncalc.calculate(atoms)\n'

        modelstr = add_relaxations(modelstr)
        return modelstr

    def get_nbands(self, n_empty=5):
        """ get the number of bands from structure, based on the number of
        valence electrons listed in utils/valence.py"""

        N_val = 0
        for symbol in self.all_symbols:
            if symbol in self.setups:
                setup = self.setups[symbol]
            else:
                setup = 's'
            N_val += VaspValence.__dict__.get(symbol)[setup]

        nbands = int(N_val / 2 + self.natoms / 2 + n_empty)
        nbands += nbands % self.ncpus

        return nbands

    def add_initial_magmoms(self, modelstr):
        modelstr += \
            """initial_magmoms = {}
symbols = atoms.get_chemical_symbols()
initial_magmom_atoms = [initial_magmoms.get(sym, 0) for sym in symbols]
atoms.set_initial_magnetic_moments(initial_magmom_atoms)\n"""\
    .format(self.initial_magmoms)
        return modelstr

    def add_calc(self, modelstr):
        modelstr += '\ncalc = Vasp(\n'
        modelstr += '    setups=setups,\n'
        for i, param in enumerate(self.sorted_parameters):
            if param == 'encut':
                continue  # Relax at ENMAX
            modelstr += '    {}={},\n'.format(param, param)

        modelstr += ')\n\ncalc.calculate(atoms)\n'
        return modelstr


def get_poscar_from_atoms(atoms):
    poscar = io.StringIO()
    write_vasp(file=poscar, atoms=atoms, vasp5=True,
               long_format=False, direct=True)

    return poscar.getvalue()


def get_model_header():
    modelstr = \
        """#!/usr/bin/env python
import os
import numpy as np
from ase.io import read
from ase.calculators.vasp import Vasp

atoms = read('initial.POSCAR')
"""
    return modelstr


def add_singlepoint(modelstr):
    modelstr += \
        """
atoms = read('OUTCAR')

for file in ['INCAR', 'OUTCAR', 'out']:
    os.rename('{}'.format(file), '{}.relax'.format(file))

if os.path.isfile('err'):
    os.rename('err', 'err.relax')
calc.set(nsw=0)

calc.calculate(atoms)
"""
    return modelstr


def add_relaxations(modelstr):
    modelstr += \
        """
atoms = read('OUTCAR', ':')

cell_change = atoms[-1].cell - atoms[0].cell

n = 1
while not np.isclose(cell_change, 0, rtol=1e-3).all():
    for file in ['INCAR', 'OUTCAR', 'out']:
        if os.path.isfile(file):
            os.rename('{}'.format(file), '{}.relax{}'.format(file, n))

    if os.path.isfile('err'):
        os.rename('err', 'err.relax')

    calc.calculate(atoms[-1])

    atoms = read('OUTCAR', ':')
    cell_change = atoms[-1].cell - atoms[0].cell
    n += 1
"""
    return modelstr



"""Pymatgen coordination environment"""

from pymatgen.core.structure import Structure
from pymatgen.io.ase import AseAtomsAdaptor
from pymatgen.analysis.chemenv.coordination_environments.\
    coordination_geometry_finder import LocalGeometryFinder


def get_neighbors_set(atoms):
    LocalGeometryFinder = LocalGeometryFinder()
    strategy = MultiWeightsChemenvStrategy.stats_article_weights_parameters()
    
    structure = AseAtomsAdaptor.get_structure(atoms)
    
    LocalGeometryFinder.setup_structure(structure=structure)

    StructureEnv = LocalGeometryFinder.compute_structure_environments(
        maximum_distance_factor=1.41,
        only_cations=False)
    
    LightStructureEnv = LightStructureEnvironments.from_structure_environments(
        strategy=strategy, structure_environments=StructureEnv)

    neighbors_set = LightStructureEnv.as_dict()['neighbors_sets']

    return neighbors_set



from setuptools import setup, find_namespace_packages

DESCRIPTION = "Software for enumerating crystal structure prototypes to by used for active " \
              "learning exploration with DFT and machine learning."

LONG_DESCRIPTION = """
Protosearch is software that supports the creation of libraries of
crystal structures according to structure prototypes derived from
the ICSD and OQMD
"""

setup(
    name='protosearch',
    url="https://github.com/ToyotaReseachInstitute/protosearch",
    version="2020.5.10",
    packages=find_namespace_packages(),
    description=DESCRIPTION,
    long_description=LONG_DESCRIPTION,
    long_description_content_type='text/markdown',
    install_requires=["ase>=3.19",
                      "numpy>=1.18",
                      "CatLearn>=0.6.2",
                      "shapely"],
    package_data={
        "protosearch.oqmd": ["*.csv"],
        "protosearch.build_bulk": ["*.db", "*.dat"],
    },
    include_package_data=True,
    extras_require={
        "enumerator": ["BulkEnumerator>=0.0.2"],
    },
    classifiers=[
          "Programming Language :: Python :: 3",
          "License :: OSI Approved :: Apache Software License",
          "Operating System :: OS Independent",
    ],
    # TODO: Edit as needed
    author="Kirsten Winther, SUNCAT-Center, TRI-AMDD",
    author_email="",
    maintainer="Kirsten Winther",
    maintainer_email="",
    license="Apache",
    keywords=[
        "materials", "chemistry", "science", "crystal",
        "density functional theory", "energy", "AI", "artificial intelligence",
        "sequential learning", "active learning"
    ],
    )


__maintainer__ = "Kirsten Winther"

import os

from protosearch import __version__ as version


def get_tri_basepath(calculator='vasp',
                 tri_path=None,
                 username=None,
                 ext=None):

    TRI_PATH = tri_path or os.environ['TRI_PATH']
    username = username or os.environ['TRI_USERNAME']

    # Protosearch version identification
    protosearch_id = 'protosearch' + version.replace('.', '')

    basepath = TRI_PATH + '/model/{}/1/u/{}/{}'\
        .format(calculator, username, protosearch_id)
    if not os.path.isdir(basepath):
        os.mkdir(basepath)
    if ext:
        basepath += '/{}'.format(ext)
        if not os.path.isdir(basepath):
            os.mkdir(basepath)

    return basepath



import os
import shlex
import subprocess
from ase.io import write

from protosearch.utils import get_tri_basepath
from protosearch.build_bulk.classification import PrototypeClassification
from .calculator import get_calculator
from .vasp import get_poscar_from_atoms

def get_submitter(cluster):
    if cluster == 'tri':
        return TriSubmit
    elif cluster == 'nersc':
        return NerscSubmit


class Submit:
    """
    Base class for writing job submission files for crystallographic
    prototype structures
    """

    def __init__(self,
                 atoms,
                 basepath,
                 calc_parameters=None,
                 ):

        self.atoms = atoms

        PC = PrototypeClassification(self.atoms)
        prototype, cell_parameters = PC.get_classification()

        self.spacegroup = prototype['spacegroup']
        self.wyckoffs = prototype['wyckoffs']
        self.species = prototype['species']

        self.cell_param_list = []
        self.cell_value_list = []
        for param in sorted(cell_parameters):
            value = cell_parameters[param]
            if value in self.cell_value_list or value in [90, 120]:
                continue
            self.cell_value_list += [cell_parameters[param]]
            self.cell_param_list += [param]

        self.basepath = basepath
        self.calculator = 'vasp'
        self.master_parameters = calc_parameters

    def write_submission_files(self):
        self.write_poscar(self.excpath)
        self.write_model(self.excpath)
        # self.write_parameters(self.excpath)

    def write_poscar(self, filepath):
        """Write POSCAR to specified file"""
        write(filepath + '/initial.POSCAR', images=self.atoms, format='vasp')

    def set_execution_path(self, strict_format=True):
        """Create a unique submission path for each structure"""

        # specify prototype for species
        path_ext = [str(self.spacegroup)]
        # wyckoffs at position
        species_wyckoffs_id = ''
        for spec, wy_spec in zip(self.species, self.wyckoffs):
            species_wyckoffs_id += spec + wy_spec
        path_ext += [species_wyckoffs_id]
        # cell parameters
        cell_param_id = ''
        if len(self.cell_param_list) < 10:
            for cell_key, cell_value in zip(self.cell_param_list,
                                            self.cell_value_list):
                param_str = '{}{}'.format(cell_key, round(cell_value, 2)).\
                    replace('c/a', 'c').replace('b/a', 'b')
                if strict_format:
                    param_str = param_str.replace('.', 'D').replace('-', 'M')
                cell_param_id += param_str

        path_ext += [cell_param_id]

        self.excroot = self.basepath
        for ext in path_ext:
            self.excroot += '/{}'.format(ext)
            if not os.path.isdir(self.excroot):
                os.mkdir(self.excroot)

        calc_revision = 1
        path_exists = True
        while os.path.isdir('{}/_{}'.format(self.excroot, calc_revision)):
            calc_revision += 1

        self.excpath = '{}/_{}'.format(self.excroot, calc_revision)
        os.mkdir(self.excpath)

    def get_calculator(self):
        symbols = self.atoms.symbols
        Calculator = get_calculator(self.calculator)
        return Calculator(symbols,
                          self.master_parameters,
                          # self.ncpus
                          )

    def write_model(self, filepath):
        """ Write model.py"""
        import json
        calculator = self.get_calculator()
        modelstr = calculator.get_model()
        parameters = calculator.calc_parameters

        with open(filepath + '/model.py', 'w') as f:
            f.write(modelstr)
        with open(filepath + '/parameters.json', 'w') as f:
            json.dump(parameters, f)

    # def write_parameters(self, filepath):
    #    with open(filepath + '/parameters.json', 'w') as f:
    #        f.write(self.calc_parameters)


class TriSubmit(Submit):
    """
    Set up (VASP) calculations on TRI-AWS for bulk structure enumerated
    with the Bulk prototype enumerator developed by A. Jain, described in:
    A. Jain and T. Bligaard, Phys. Rev. B 98, 214112 (2018)

    Parameters:

    atoms: ASE Atoms object
    calc_parameters: dict
        Optional specification of parameters, such as {ecut: 300}.
        If not specified, the parameter standards given in
        ./utils/standards.py will be applied
    ncpus: int
        number of cpus on AWS to use, default: 1
    queue: 'small', 'medium', etc
        Queue specificatin for AWS
    calculator: str
        Currently only 'vaps' is implemented
    basepath: str or None
        Path for job submission of in TRI filesync (s3) directory
        F.ex: '~/matr.io//model/<calculator>/1/u/<username>

    Alternative to setting basepath, use the following environment valiables
    to set the job submission path automatically:
        TRI_PATH: Your TRI sync directory, which is usually at ~/matr.io
        TRI_USERNAME: Your TRI username
    """

    def __init__(self,
                 atoms,
                 calc_parameters=None,
                 ncpus=None,
                 queue='medium',
                 calculator='vasp',
                 basepath=None,
                 basepath_ext=None
                 ):

        if basepath is None:
            basepath = get_tri_basepath(calculator=calculator,
                                        ext=basepath_ext)

        super().__init__(atoms,
                         basepath,
                         calc_parameters)

        if ncpus is None:
            ncpus = get_ncpus_from_volume(atoms)

        self.ncpus = ncpus
        self.queue = queue

    def submit_calculation(self):
        """Submit calculation for unique structure.
        First the execution path is set, then the initial POSCAR and models.py
        are written to the directory.

        The calculation is submitted as a regular model with trisub.
        """

        self.set_execution_path()
        self.write_submission_files()

        command = shlex.split('trisub -q {} -c {}'.format(
            self.queue, self.ncpus))
        subprocess.call(command, cwd=self.excpath)


class NerscSubmit(Submit):
    """
    WIP!

    Set up VASP calculations on NERSC for bulk structure enumerated
    with the Bulk prototype enumerator developed by A. Jain, described in:
    A. Jain and T. Bligaard, Phys. Rev. B 98, 214112 (2018)

    atoms: ASE Atoms object
    calc_parameters: dict
        Optional specification of parameters, such as {ecut: 300}.
        If not specified, the parameter standards given in
        ./utils/standards.py will be applied
    qos: quality of service SLURM option
       'debug', 'scavenger' or 'premium'
    partition: SLURM partition
       'regular' or 'scavenger'
    time: SLURM job time allocation
    nodes: SLURM node allocation
    ntasks: number of processes per node. 66 on cori knl  
    basepath: str
        Path to your root project directory for running calculations
        defaults to $SCRATCH/protosearch
    """

    def __init__(self,
                 atoms,
                 account,
                 calc_parameters=None,
                 qos='premium',
                 partition='regular',
                 time='1.00.00',
                 nodes=10,
                 ntasks=66,
                 basepath='$SCRATCH/protosearch',
                 ):

        super().__init__(atoms,
                         basepath,
                         calc_parameters)

        self.qos = qos
        self.partition = partition
        self.time = time
        self.account = account
        self.nodes = nodes
        self.ntasks = ntasks

    def submit_calculation(self):
        """Submit calculation for structure
        First the execution path is set, then the initial POSCAR and models.py
        are written to the directory.
        """

        self.set_execution_path(strict_format=False)
        self.write_submission_files()
        self.write_submit_script()

        command = shlex.split('sh submit.sh')
        subprocess.call(command, cwd=self.excpath)

    def write_submit_script(self):
        script = get_nersc_submit_script(self,
                                         self.excpath,
                                         qos=self.qos,
                                         partition=self.partition,
                                         time=self.time,
                                         account=self.account,
                                         nodes=self.nodes,
                                         ntasks=self.ntasks)
        with open(self.excpath + '/submit.sh', 'w') as f:
            f.write(script)


def get_ncpus_from_volume(atoms):
    ncpus = int((atoms.get_volume() / 80 // 4) * 4)
    return max(ncpus, 1)


def get_nersc_submit_script(self,
                            excpath,
                            account,
                            qos='premium',
                            partition='regular',
                            time='1:00:00',
                            nodes=10,
                            ntasks=66):
    """Write SLURM submit script for NERSC. Takes regular SLURM parameters
    qos (debug, scavenger, premium)
    partition (regular, scavenger)
    """

    script = '#!/bin/bash\n'
    script += '#SBATCH -q {}\n'.format(qos)
    script += '#SBATCH -p {}\n'.format(partition)
    script += '#SBATCH -t {}\n'.format(time)
    script += '#SBATCH -A {}\n'.format(account)
    script += '#SBATCH -e error.txt\n'
    script += '#SBATCH -o output.txt\n'
    script += '#SBATCH -C knl\n'
    script += '#SBATCH --ntasks-per-node={}\n'.format(ntasks)

    script += """
module load python/3.7-anaconda-2019.07
export PYTHONPATH=$HOME/.local/cori/3.7-anaconda-2019.07/lib/python3.7/site-packages:$PYTHONPATH
module load vasp-tpc/5.4.4-knl

#cd {excpath}
export VASP_SCRIPT=./run_vasp.py
export VASP_PP_PATH=/project/projectdirs/{account}/vasp-psp/pseudo52

export OMP_NUM_THREADS=1
export OMP_PLACES=threads
export OMP_PROC_BIND=spread
""".format(excpath=excpath, account=account)

    script += """
echo "import os" > run_vasp.py
echo "exitcode = os.system('srun -n {nproc} -c 4 --cpu_bind=cores vasp_std')" >> run_vasp.py

echo $VASP_SCRIPT
echo $VASP_PP_PATH

python {excpath}/model.py""".format(nproc=nodes * ntasks, excpath=excpath)

    return script



import sys
import string
import json
import time
import numpy as np
from pathos.multiprocessing import ProcessingPool as Pool
import bulk_enumerator as be
from ase.symbols import string2symbols

from protosearch.build_bulk.build_bulk import BuildBulk
from protosearch.build_bulk.cell_parameters import CellParameters
from protosearch.workflow.prototype_db import PrototypeSQL
from .fitness_function import get_covalent_density, get_fitness


class Enumeration():

    def __init__(self,
                 stoichiometry,
                 num_start,
                 num_end,
                 SG_start=1,
                 SG_end=230,
                 num_type='atom'):
        """
        Parameters

        stoichiometry: str
            Ratio bewteen elements separated by '_'.
            For example: '1_2' or '1_2_3'

        num_start: int
            minimum number of atoms or wyckoff sites
        num_end: int
            maximum number of atoms or wyckoff sites
        SG_start: int
           minimum spacegroup number
        SG_end: int
           maximum spacegroup number
        num_type: str
            'atom' or 'wyckoff'
        """

        self.stoichiometry = stoichiometry
        self.num_start = num_start
        self.num_end = num_end
        self.SG_start = SG_start
        self.SG_end = SG_end
        self.num_type = num_type

    def set_spacegroup(self, spacegroup):
        self.SG_start = spacegroup
        self.SG_end = spacegroup + 1

    def set_stoichiometry(self, stoichiometry):
        self.stoichiometry = stoichiometry

    def set_natoms(self, natoms):
        self.num_start = natoms
        self.num_end = natoms

    def get_enumeration(self):
        E = be.enumerator.ENUMERATOR()
        enumerations = E.get_bulk_enumerations(self.stoichiometry,
                                               self.num_start,
                                               self.num_end,
                                               self.SG_start,
                                               self.SG_end,
                                               self.num_type)

        return enumerations

    def store_enumeration(self, filename=None):
        for SG in range(self.SG_start, self.SG_end + 1):
            with PrototypeSQL(filename=filename) as DB:
                enumerated = [DB.is_enumerated(self.stoichiometry,
                                               SG, num, self.num_type)
                              for num in range(self.num_start, self.num_end + 1)]
                if np.all(enumerated):
                    print('spacegroup={} already enumerated for n_{}={}:{}'
                          .format(SG, self.num_type,
                                  self.num_start, self.num_end))
                    continue
            E = be.enumerator.ENUMERATOR()
            try:
                enumerations = E.get_bulk_enumerations(
                    self.stoichiometry,
                    self.num_start,
                    self.num_end,
                    SG,
                    SG,
                    self.num_type)

                print('Found {} prototypes for spacegroup={}'.format(
                    len(enumerations), SG))
            except:
                print('Found 0 prototypes for spacegroup={}'.format(SG))

            with PrototypeSQL(filename=filename) as DB:
                for entry in enumerations:
                    entry.update({'source': 'prototype'})
                    DB.write_prototype(entry=entry)

                for num in range(self.num_start, self.num_end + 1):
                    DB.write_enumerated(self.stoichiometry, SG, num,
                                        self.num_type)


all_elements = [
    'H', 'He',
    'Li', 'Be', 'B', 'C', 'N', 'O', 'F', 'Ne',
    'Na', 'Mg', 'Al', 'Si', 'P', 'S', 'Cl', 'Ar',
    'K', 'Ca', 'Sc', 'Ti', 'V', 'Cr', 'Mn', 'Fe', 'Co', 'Ni', 'Cu', 'Zn',
    'Ga', 'Ge', 'As', 'Se', 'Br', 'Kr',
    'Rb', 'Sr', 'Y', 'Zr', 'Nb', 'Mo', 'Tc', 'Ru', 'Rh', 'Pd', 'Ag', 'Cd',
    'In', 'Sn', 'Sb', 'Te', 'I', 'Xe',
    'Cs', 'Ba', 'La', 'Ce', 'Pr', 'Nd', 'Pm', 'Sm', 'Eu', 'Gd', 'Tb', 'Dy',
    'Ho', 'Er', 'Tm', 'Yb', 'Lu',
    'Hf', 'Ta', 'W', 'Re', 'Os', 'Ir', 'Pt', 'Au', 'Hg', 'Tl', 'Pb', 'Bi',
    'Po', 'At', 'Rn']

non_metals = ['H', 'He', 'B', 'C', 'N', 'O', 'F', 'Ne',
              'Si', 'P', 'S', 'Cl', 'Ar',
              'Ge', 'As', 'Se', 'Br', 'Kr',
              'Sb', 'Te', 'I', 'Xe',
              'Po', 'At', 'Rn']

metals = [
    'Li', 'Be',
    'Na', 'Mg', 'Al',
    'K', 'Ca', 'Sc', 'Ti', 'V', 'Cr', 'Mn', 'Fe', 'Co', 'Ni', 'Cu', 'Zn',
    'Ga',
    'Rb', 'Sr', 'Y', 'Zr', 'Nb', 'Mo', 'Tc', 'Ru', 'Rh', 'Pd', 'Ag', 'Cd',
    'In', 'Sn',
    'Cs', 'Ba', 'La', 'Ce', 'Pr', 'Nd', 'Pm', 'Sm', 'Eu', 'Gd', 'Tb', 'Dy',
    'Ho', 'Er', 'Tm', 'Yb', 'Lu',
    'Hf', 'Ta', 'W', 'Re', 'Os', 'Ir', 'Pt', 'Au', 'Hg', 'Tl', 'Pb', 'Bi',
]


class AtomsEnumeration():
    """
    Enumeratate atomic structures based on prototype and elements

    elements: dict
        {'A': 'metals', 'B': ['O', 'F'], 'C': 'all'}
    """

    def __init__(self,
                 elements,
                 max_atoms=None,
                 spacegroups=None):
        self.elements = elements
        self.max_atoms = max_atoms
        self.spacegroups = spacegroups or list(range(1, 231))

        for key, value in self.elements.items():
            self.elements[key] = map_elements(value)

    def store_atom_enumeration(self, filename=None, multithread=False):
        self.filename = filename
        DB = PrototypeSQL(filename=filename)
        DB._connect()
        N0 = DB.ase_db.count()

        prototypes = DB.select(max_atoms=self.max_atoms,
                               spacegroups=self.spacegroups,
                               source='prototype')
        Nprot = len(prototypes)

        pool = Pool()

        t0 = time.time()
        if multithread:
            res = pool.amap(self.store_atoms_for_prototype, prototypes)
            while not res.ready():
                N = DB.ase_db.count() - N0
                t = time.time() - t0
                N_per_t = N / t
                if N > 0:
                    print('---------------------------------')
                    print(
                        "{}/{} structures generated in {:.2f} sec".format(N, Nprot, t))
                    print("{} sec / structure".format(t / N))
                    print('Estimated time left: {:.2f} min'.format(
                        Nprot / N_per_t / 60))
                print('---------------------------------')
                time.sleep(10)
            res = res.get()
        else:
            for prototype in prototypes:
                self.store_atoms_for_prototype(prototype)

    def store_atoms_for_prototype(self, prototype, max_candidates=3):

        p_name = prototype['name']
        counts = []
        for a in p_name.split('_')[0]:
            if a.isdigit():
                counts[-1] += int(a) - 1
            else:
                counts += [1]

        species_lists = self.get_species_lists(
            prototype['species'], prototype['permutations'], counts)

        cell_parameters = prototype.get('cell_parameters', None)
        if cell_parameters:
            cell_parameters = json.load(cell_parameters)
        for species in species_lists:
            structure_name = str(prototype['spacegroup'])
            for spec, wy_spec in zip(species, prototype['wyckoffs']):
                structure_name += '_{}_{}'.format(spec, wy_spec)
            with PrototypeSQL(filename=self.filename) as DB:
                if DB.ase_db.count(structure_name=structure_name) > 0:
                    continue

                for row in DB.ase_db.select(p_name=prototype['name'], limit=1):
                    cell_parameters = json.loads(row.cell_parameters)

            BB = BuildBulk(prototype['spacegroup'],
                           prototype['wyckoffs'],
                           species,
                           )
            atoms_list, parameters = \
                BB.get_wyckoff_candidate_atoms(proximity=1,
                                               primitive_cell=True,
                                               return_parameters=True,
                                               max_candidates=max_candidates)

            key_value_pairs = {'p_name': prototype['name'],
                               'spacegroup': prototype['spacegroup'],
                               'wyckoffs':
                               json.dumps(prototype['wyckoffs']),
                               'species': json.dumps(species),
                               'structure_name': structure_name,
                               'relaxed': 0,
                               'completed': 0,
                               'submitted': 0}

            for i, atoms in enumerate(atoms_list):
                atoms.info.pop('spacegroup')
                if 'spacegroup_kinds' in atoms.info:
                    atoms.info.pop('spacegroup_kinds')
                key_value_pairs.update(atoms.info)
                key_value_pairs.update(
                    {'cell_parameters': json.dumps(parameters[i])})

                fitness = get_fitness(atoms)
                apf = get_covalent_density(atoms)

                key_value_pairs.update({'fitness': fitness,
                                        'apf': apf})

                with PrototypeSQL(filename=self.filename) as DB:
                    DB.ase_db.write(atoms, key_value_pairs)

    def get_species_lists(self, gen_species, permutations, counts):

        elements = self.elements

        alph = list(string.ascii_uppercase)

        name_map = {'A': 'A0',
                    'B': 'A1',
                    'C': 'A2',
                    'D': 'A3',
                    'E': 'A4',
                    'F': 'A5',
                    'G': 'A6',
                    'G': 'A7'}

        N_unique = len(set(gen_species))

        N_species = len(list(elements.keys()))

        c_list = [[]]
        for n in range(N_unique):
            c_list_0 = c_list.copy()
            c_list = []
            for c in c_list_0:
                for e in elements[alph[n]]:
                    if not e in c:
                        c_list += [c + [e]]

        i = 0
        group_species = [[0]]
        while i < len(counts) - 1:
            if counts[i] == counts[i + 1]:
                group_species[-1] += [i + 1]
            else:
                group_species += [[i + 1]]
            i += 1

        # Only include permutations between same atomic ratio

        permutations_list = [p.split('_') for p in permutations]

        permutations_list = [[int(p[1]) for p in p_list] for
                             p_list in permutations_list]
        for g in group_species:
            if len(g) == 1:
                i = g[0]
                permutations_list = [p for p in permutations_list if
                                     p[i] == i]
            elif len(g) > 1:
                sub_perm = np.sort(g)
                permutations_list = [p for p in permutations_list if
                                     np.all(np.sort(np.array(p)[g]) == sub_perm)]

        extra_c_list = []
        for c in c_list:
            for p_list in permutations_list[1:]:
                new = list(np.array(c)[p_list])
                extra_c_list += [new]

        c_list += extra_c_list

        species_list = []

        for c in c_list:
            temp = []
            for g in gen_species:
                for i, cc in enumerate(c):
                    g = g.replace(name_map[alph[i]], cc)
                temp += [g]
            species_list += [temp]

        return species_list


def map_elements(key):
    if isinstance(key, list):
        assert np.all([k in all_elements for k in key])
        return key

    key_to_elements = {'metals': metals,
                       'non_metals': non_metals,
                       'all': all_elements}

    return key_to_elements[key]


def get_stoich_from_formulas(formulas):
    alphabet = ['A', 'B', 'C', 'D', 'E', 'F', 'G', 'H']
    stoichs = []
    elements = {}
    for formula in formulas:
        stoich, unique_symbols = formula2stoich(formula)
        stoichs += [stoich]
        for i, us in enumerate(unique_symbols):
            if not alphabet[i] in elements:
                elements[alphabet[i]] = []
            elements[alphabet[i]] += [us]
    stoichs = list(set(stoichs))
    return stoichs, elements


def formula2stoich(formula):
    symbols = string2symbols(formula)
    unique_symbols = list(set(symbols))
    count = [symbols.count(us) for us in unique_symbols]
    idx = np.argsort(count)
    count = [count[i] for i in idx]
    unique_symbols = [unique_symbols[i] for i in idx]
    stoich = '_'.join([str(c) for c in count])
    return stoich, unique_symbols


def get_formulas(elements,
                 stoichiometries=None,
                 max_atoms=12):
    """
    obtain all possible formulas for a
    supplied set of elements.

    stoichiometries:  list or None
        f.ex. ['1_2', '1_3', '1_2_2']
    max_atoms: int
    """

    if not stoichiometries:
        """Get all possible combinations where N_A <= N_B etc. and
        sum(N) <= max_atoms"""

        N_species = len(list(elements.keys()))
        n_max = max_atoms - N_species
        s_dict = {'0': [[i] for i in range(1, n_max // N_species + 1)]}
        dim = 1
        while dim < N_species:
            s_dict.update({str(dim): []})
            for s_list in s_dict[str(dim - 1)]:
                n_atoms = sum(s_list)
                i = s_list[-1]
                for j in [j for j in range(1, max_atoms) if j >= i]:
                    if n_atoms + j > max_atoms:
                        continue
                    s_dict[str(dim)] += [s_list + [j]]
            dim += 1

        stoichiometries = []
        for s_list in s_dict[str(dim - 1)]:
            stoichiometries += ['_'.join([str(s) for s in s_list])]

    alph = list(string.ascii_uppercase)

    all_formulas = []
    for stoichiometry in stoichiometries:
        assert len(elements.keys()) == \
            len(stoichiometry.split('_'))
        formulas = np.array([])
        for i, value in enumerate(stoichiometry.split('_')):
            element_list = elements[alph[i]]
            if value == '1':
                value = ''
            element_list = np.char.array([e + value for e in element_list])
            if len(formulas) == 0:
                formulas = element_list
            else:
                formulas = np.expand_dims(formulas, axis=i).\
                    repeat(len(element_list), axis=i)
                for k in range(i):
                    append_formulas = np.expand_dims(element_list, axis=k).\
                        repeat(formulas.shape[k], axis=k)
                formulas = formulas + append_formulas
        all_formulas += list(formulas.flatten())
    return all_formulas
